import logging
import uuid
from typing import List, Dict

from atscale.errors import atscale_errors
from atscale.base import enums, templates
from atscale.parsers import project_parser
from atscale.utils import dimension_utils, measure_utils, project_utils, time_utils
from atscale.db.sql_connection import SQLConnection
from atscale.connection import Connection

logger = logging.getLogger(__name__)


def add_data_set_ref(model_dict, dataset_id):
    data_set_ref = templates.create_dataset_ref_dict(dataset_id)
    model_dict.setdefault('data-sets', {})
    model_dict['data-sets'].setdefault('data-set-ref', [])
    model_dict['data-sets']['data-set-ref'].append(data_set_ref)


def check_features(features: list, check_list: list, errmsg: str = None):
    """Checks that the given feature(s) exist(s) within a specified list of features.
    Args:
        features (list): feature(s) to confirm exist in the provided list
        check_list (list): features of the data model to check against
        errmsg (str, optional): Error message to raise if feature not found. Defaults to None.
    Raises:
        atscale_errors.UserError: error is raised if any featuers of interest not found, can be customized via errmsg arg
    Returns:
        boolean: True if no error found
    """
    setDif = set(features) - set(check_list)
    if len(setDif) > 0:
        if errmsg:
            raise atscale_errors.UserError(errmsg)
        else:
            raise atscale_errors.UserError(
                f'Features: {sorted(list(setDif))} not in data model. Make sure each feature has been published and is'
                ' correctly spelled')
    return True


def create_dataset_relationship(atconn, project_dict, cube_id, database, schema, table_name, join_features,
                                join_columns=None,
                                roleplay_features=None):
    """ Mutates and returns the given project_dict to create a dataset, join the given features, and join the dataset
    to the cube if it was not already.

    Args:
        atconn (Connection): A Connection object connected to the server of the project the parameters correspond to.
        project_dict (dict): The project_dict to mutate and return
        cube_id (str): The id of the cube the dataset will be joined to.
        database (str): The database that the created dataset will point at.
        schema (str): The schema that the created dataset will point at.
        table_name (str): The table that the created dataset will point at.
            This will also become the name of the dataset.
        join_features (list): a list of features in the data model to use for joining.
        join_columns (list, optional): The columns in the dataframe to join to the join_features. List must be either
            None or the same length and order as join_features. Defaults to None to use identical names to the
            join_features. If multiple columns are needed for a single join they should be in a nested list
        roleplay_features (list, optional): The roleplays to use on the relationships. List must be either
            None or the same length and order as join_features. Use '' to not roleplay that relationship.
            Defaults to None.

    Returns:
        Dict: The mutated project_dict
        """
    warehouse_id = project_utils.get_project_warehouse(project_dict)
    project_datasets = project_parser.get_datasets(project_dict)

    if join_columns is None:
        join_columns = join_features

    if roleplay_features is None:
        roleplay_features = []
        for feature in join_features:
            roleplay_features.append('')

    if len(join_features) != len(join_columns):
        raise Exception(f'join_features and join_columns lengths must match. join_features is length '
                        f'{len(join_features)} while join_columns is length {len(join_columns)}')

    if roleplay_features is not None and len(join_features) != len(roleplay_features):
        raise Exception(f'join_features and roleplay_features lengths must match. join_features is length '
                        f'{len(join_features)} while roleplay_features is length {len(roleplay_features)}')

    url = f'{atconn._warehouse_endpoint()}/conn/{warehouse_id}/tables/cacheRefresh'
    atconn._submit_request(request_type=enums.RequestType.POST, url=url)

    # we'll use table_columns potentially for creating a dataset below and then more logic after that
    table_columns = atconn.get_table_columns(warehouse_id=warehouse_id,
                                             table_name=table_name,
                                             database=database,
                                             schema=schema)

    # there was a check as to whether the datasets element was null which set it to an empty list. I'm not sure if that would
    # ever be possible. If so, I imagine project_datasets would have to be null, and then verify_connection would fail. If we
    # want to support this, then we'd have to change verify to not check project_data set values. I rearranged the logic a lot.
    # Figured that was safe to pull out into an independent check here.
    project_dict['datasets'].setdefault('data-set', [])

    # look for a dataset that may already have the table_name for the table we're trying to join to the cube (meaning the table
    # already exists and we're just replacing it or appending it)
    dataset_id = project_parser.find_dataset_with_table(
        project_datasets, table_name)

    if not dataset_id:  # then we have to create a project_dataset
        # the prior code assumed a schema but checked if database was None prior to setting
        project_dataset, dataset_id = project_utils.create_dataset(
            table_name, warehouse_id, table_columns, database, schema)
        # and we'll add the newly minted dataset to the project_dict
        project_dict['datasets']['data-set'].append(project_dataset)

    return create_dataset_relationship_from_dataset(project_dict=project_dict,
                                                    cube_id=cube_id,
                                                    dataset_name=table_name,
                                                    join_features=join_features,
                                                    join_columns=join_columns,
                                                    roleplay_features=roleplay_features)


def create_dataset_relationship_from_dataset(project_dict: dict,
                                             cube_id: str,
                                             dataset_name: str,
                                             join_features: List[str],
                                             join_columns: List[str],
                                             roleplay_features: List[str]):
    """ Mutates and returns the given project_dict to join the given features and join the dataset, of the given name,
    to the cube (if not joined already) of the given cube_id. 

    Args:
        atconn (Connection): A Connection object connected to the server of the project the parameters correspond to.
        project_dict (dict): The project_dict to mutate and return.
        cube_id (str): The id of the cube the dataset will be joined to.
        dataset_name (str): The name of the dataset to target. This dataset must exist in project_datasets.
            This will also become the name of the dataset.
        join_features (list): a list of features in the data model to use for joining.
        join_columns (list, optional): The columns in the dataframe to join to the join_features. List must be either
            None or the same length and order as join_features. Defaults to None to use identical names to the
            join_features. If multiple columns are needed for a single join they should be in a nested list
        roleplay_features (list, optional): The roleplays to use on the relationships. List must be either
            None or the same length and order as join_features. Use '' to not roleplay that relationship. 
            Defaults to None.

    Returns:
        Dict: The mutated project_dict
        """
    if join_columns is None:
        join_columns = join_features
    elif len(join_features) != len(join_columns):
        raise Exception(f'join_features and join_columns lengths must match. join_features is length '
                        f'{len(join_features)} while join_columns is length {len(join_columns)}')

    if roleplay_features is None:
        # [''] * len() will make references to the same str
        roleplay_features = ['' for x in range(len(join_features))]
    elif roleplay_features is not None and len(join_features) != len(roleplay_features):
        raise Exception(f'join_features and roleplay_features lengths must match. join_features is length '
                        f'{len(join_features)} while roleplay_features is length {len(roleplay_features)}')

    project_datasets = project_parser.get_datasets(project_dict)
    dataset_id = project_parser.get_dataset_from_datasets_by_name(
        project_datasets=project_datasets,
        dataset_name=dataset_name)['id']

    key_refs = []
    attribute_refs = []

    joins = tuple(zip(join_features, join_columns, roleplay_features))
    cube_dict = project_parser.get_cube(project_dict, cube_id)

    # create dict from all feature name to their index in project level attributes
    feat_name_to_project_attribute: Dict[str, int] = {}
    project_keyed_attributes = project_dict.get(
        'attributes', {}).get('keyed-attribute', [])
    for index, attribute in enumerate(project_keyed_attributes):
        feat_name_to_project_attribute[attribute['name']] = index

    for join_feature, join_column, roleplay_feature in joins:
        # Questions :
        # Why are we making a list for each join feature, column, and roleplay feature?
        # Cols is a list and we set it on key_ref - can that literally be more than on column? Why is it a list? If it is a list, why is the key not cols instead of col?
        if type(join_column) != list:
            # method can take a list of [join_column] or just a list of strings
            join_column = [join_column]

        # looks in the project for an attribute that matches the join feature. If one is found,
        # it determines if user specified it as a roleplay feature. If they did, it looks for the name
        # at the current location to see if it matches the name provided by the user for the roleplay feature
        # and if that is not found, it appends the current name in place, to the provided roleplay  feature,
        # and constructs the json around it, like ref_id, and sets it up to reference the dimension id
        dimension_index = feat_name_to_project_attribute.get(join_feature)
        if dimension_index is not None:  # don't do bool because if index is 0 it will return false
            dimension = project_keyed_attributes[dimension_index]
            ref = dimension['key-ref']
            key_ref = {
                'id': ref,
                'unique': False,
                'complete': 'false',
                'column': join_column
            }
            if roleplay_feature != '':
                if '{0}' not in roleplay_feature:
                    roleplay_feature = roleplay_feature + ' {0}'
                ref_id = str(uuid.uuid4())
                ref_path = {
                    'new-ref': {
                        'attribute-id': dimension['id'],
                        'ref-id': ref_id,
                        'ref-naming': roleplay_feature
                    }
                }
                key_ref['ref-path'] = ref_path
            key_refs.append(key_ref)

        # if the join features was not found as a dimension in project_json above, look in the cube.
        if not dimension_index:
            cube_keyed_attributes = cube_dict.get(
                'attributes', {}).get('keyed-attribute', [])
            feat_name_to_cube_attribute: Dict[str, int] = {
                x['name']: i for i, x in enumerate(cube_keyed_attributes)}
            dimension_index = feat_name_to_cube_attribute.get(join_feature)
            if dimension_index is not None:
                dimension = cube_keyed_attributes[dimension_index]
                ref = dimension['key-ref']
                key_ref = {
                    'id': ref,
                    'unique': False,
                    'complete': 'partial',
                    'column': join_column
                }
                if roleplay_feature != '':
                    if '{0}' not in roleplay_feature:
                        roleplay_feature = roleplay_feature + ' {0}'
                    ref_id = str(uuid.uuid4())
                    ref_path = {
                        'new-ref': {
                            'attribute-id': dimension['id'],
                            'ref-id': ref_id,
                            'ref-naming': roleplay_feature
                        }
                    }
                    key_ref['ref-path'] = ref_path
                key_refs.append(key_ref)
                uid = dimension['id']
                attr = {
                    'id': uid,
                    'complete': 'partial',
                    'column': join_column
                }
                attribute_refs.append(attr)
    found = False
    cube_dict.setdefault('data-sets', {})
    cube_dict['data-sets'].setdefault('data-set-ref', [])
    for ds_ref in cube_dict['data-sets']['data-set-ref']:
        if ds_ref['id'] == dataset_id:
            found = True
            ds_ref['logical'].setdefault('key-ref', [])
            ds_ref['logical']['key-ref'] = ds_ref['logical']['key-ref'] + key_refs
            ds_ref['logical'].setdefault('attribute-ref', [])
            ds_ref['logical']['attribute-ref'] = ds_ref['logical']['attribute-ref'] + attribute_refs
            break

    # If we had to create a dataset for the project to point at the new table, then we need to ensure there is also one in the cube referencing it.
    # This check previously referred back to "found" which was based on the project data set being there, but this is really about whether we
    # find it in the cube, which is in the logic immediately above, so I'll do the boolean there instead.
    if not found:
        data_set_ref = templates.create_dataset_ref_dict(
            dataset_id, key_refs, attribute_refs)
        cube_dict['data-sets']['data-set-ref'].append(data_set_ref)

    return project_dict

def create_semantic_layer(atconn: Connection, dbconn: SQLConnection, table_name: str, project_dict: dict, data_model_id: str, dataset_id: str, columns: list):
    """Mutates the provided project_dict to add a semantic layer. NOTE: This does not update the project! Calling methods must still update and publish the project using the resulting project_dict. 

    Args:
        atconn (Connection): AtScale connection
        dbconn (Connection): DB connection
        table_name (str): the name of the table to create a semantic table for
        project_dict (dict): the project dictionary (generally sparse result of creating a new project and adding a dataset)
        data_model_id (str): the id for the data_model (generally sparse result of creating a new project)
        dataset_id (str): the id for the dataset associated with the table_name for which we will create a semantic layer
        columns (list[tuple]): columns of the table associated with table_name and dataset_id as AtScale sees them, generally with a name and type for each
    """
    for column in columns:
        column_name = column[0]
        # this does a string comparison to see if this column type is a DateTime
        if enums.AtScaleColumnTypes.DateTime.value in column[1] or enums.AtScaleColumnTypes.Date.value in column[1]:
            # Add a dimension for the date column
            try:#determine_time_levels depends on count(distinct(column_name)) sql working. If the db errors out, then we just skip
                time_levels = time_utils.determine_time_levels(
                    dbconn=dbconn, table_name=table_name, column=column_name)
            except Exception as e:
                logger.error(f"Unable to determine TimeLevels in create_semantic_layer for column {column} and db type {dbconn.platform_type}. The error was{e}")
                #skip the rest and go to the next column in the loop
                continue 
            dimension_utils.create_time_dimension_for_column(
                atconn=atconn, project_dict=project_dict, data_model_id=data_model_id, dataset_id=dataset_id, column_name=column_name, time_levels=time_levels)

        elif enums.AtScaleColumnTypes.String.value in column[1]:
            dimension_utils.create_categorical_dimension_for_column(
                atconn=atconn, project_dict=project_dict, data_model_id=data_model_id, dataset_id=dataset_id, column_name=column_name)

        # could pile on the various numeric types, not sure how AtScale sees them all, so far I've seen "Decimal" a lot.
        elif enums.AtScaleColumnTypes.Decimal.value in column[1]:
            # Note that calculated columns pulling out parts of dates will show up as Decimal data type also.
            measure_utils.create_measure(
                project_dict=project_dict, data_model_id=data_model_id, dataset_id=dataset_id, column_name=column_name)

    # The default data_model object when creating a project and writing a dataframe only has a data-set-ref. If we added dimensions above,
    # then we need to add some other dict elements to the data_model. I'm not actually sure how these are used. Just going with some defaults here
    data_model_dict = project_parser.get_cube(
        project_dict=project_dict, id=data_model_id)
    data_model_dict.setdefault(
        'properties', templates.create_data_model_properties_dict_default())
    data_model_dict.setdefault(
        'actions', templates.create_data_model_actions_dict_default())
    # this being empty seems weird since we have calculated columns, but maybe this refers to calculated measures?
    data_model_dict.setdefault('calculated-members', {})
    data_model_dict.setdefault('aggregates', {})
